"""add_login_audit_columns

Revision ID: a3e2b0a4937d
Revises: 53e97221d99f
Create Date: 2020-01-09 14:23:56.880364

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a3e2b0a4937d"
down_revision = "53e97221d99f"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("user", sa.Column("last_login_at", sa.DateTime(), nullable=True))
    op.add_column("user", sa.Column("login_count", sa.Integer, default=0))


def downgrade():
    op.drop_column("user", "last_login_at")
    op.drop_column("user", "login_count")



from typing import Dict
import importlib

from . import PluginType


def plugin_factory(plugin_type: PluginType, plugin_def: Dict):
    def lazy_import(module, cls):
        def lazy():
            return getattr(importlib.import_module(module, "meltano.core.plugin"), cls)

        return lazy

    plugin_class = {
        PluginType.EXTRACTORS: lazy_import(".singer", "SingerTap"),
        PluginType.LOADERS: lazy_import(".singer", "SingerTarget"),
        PluginType.TRANSFORMS: lazy_import(".dbt", "DbtTransformPlugin"),
        PluginType.MODELS: lazy_import(".model", "ModelPlugin"),
        PluginType.DASHBOARDS: lazy_import(".dashboard", "DashboardPlugin"),
        PluginType.ORCHESTRATORS: lazy_import(".airflow", "Airflow"),
        PluginType.TRANSFORMERS: lazy_import(".dbt", "DbtPlugin"),
    }

    # this will parse the discovery file and create an instance of the
    # corresponding `plugin_class` for all the plugins.
    plugin_cls = plugin_class[plugin_type]()

    return plugin_cls(plugin_def.pop("name"), **plugin_def)



import os
import yaml
import json
import click
import sys
import logging
from urllib.parse import urlparse

from . import cli
from .params import project
from meltano.core.project_add_service import (
    ProjectAddService,
    PluginNotSupportedException,
    PluginAlreadyAddedException,
)
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.project_add_custom_service import ProjectAddCustomService
from meltano.core.plugin_install_service import (
    PluginInstallService,
    PluginNotInstallable,
)
from meltano.core.plugin import PluginType, Plugin
from meltano.core.project import Project
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.error import SubprocessError
from meltano.core.db import project_engine


@cli.command()
@click.argument(
    "plugin_type", type=click.Choice([type.cli_command for type in list(PluginType)])
)
@click.argument("plugin_name")
@click.option("--custom", is_flag=True)
@click.option("--include-related", is_flag=True)
@project()
@click.pass_context
def add(ctx, project, plugin_type, plugin_name, **flags):
    if flags["custom"]:
        if plugin_type in ("transformer", "transform", "orchestrator"):
            click.secho(f"--custom is not supported for {ctx.invoked_subcommand}")
            raise click.Abort()

        add_service = ProjectAddCustomService(project)
    else:
        add_service = ProjectAddService(project)

    add_plugin(
        add_service,
        project,
        PluginType(f"{plugin_type}s"),
        plugin_name,
        include_related=flags["include_related"],
    )

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type=plugin_type, plugin_name=plugin_name)


def add_plugin(
    add_service,
    project: Project,
    plugin_type: PluginType,
    plugin_name: str,
    include_related=False,
):
    try:
        plugin = add_service.add(plugin_type, plugin_name)
        click.secho(f"Added '{plugin_name}' to your Meltano project.", fg="green")
    except PluginAlreadyAddedException as err:
        click.secho(
            f"'{plugin_name}' was found in your Meltano project. Use `meltano install` to install it.",
            fg="yellow",
            err=True,
        )
        plugin = err.plugin
    except (PluginNotSupportedException, PluginNotFoundError):
        click.secho(f"Error: {plugin_type} '{plugin_name}' is not supported.", fg="red")
        raise click.Abort()

    install_service = PluginInstallService(project)

    try:
        click.secho(f"Installing '{plugin_name}'...")
        run = install_service.install_plugin(plugin)
        if run:
            click.secho(run.stdout)
        click.secho(f"Installed '{plugin_name}'.", fg="green")

        click.secho(f"Added and installed {plugin_type} '{plugin_name}'.", fg="green")
    except PluginNotInstallable as install_err:
        logging.info(f"{plugin_type} is not installable, skipping install.")
    except SubprocessError as proc_err:
        click.secho(str(proc_err), fg="red")
        click.secho(proc_err.process.stderr, err=True)
        raise click.Abort()

    docs_link = plugin._extras.get("docs")
    if docs_link:
        click.secho(f"Visit {docs_link} for more details about '{plugin.name}'.")

    if include_related:
        discovery_service = PluginDiscoveryService(project)
        plugin_def = discovery_service.find_plugin(plugin.type, plugin.name)

        related_plugins = add_service.add_related(plugin_def)
        if len(related_plugins) == 0:
            click.secho("No related plugins found that are not already installed.")
        else:
            for plugin in related_plugins:
                click.secho(
                    f"Added related plugin '{plugin.name}' to your Meltano project.",
                    fg="green",
                )

            click.secho(f"Installing {len(related_plugins)} related plugins...")
            install_status = install_service.install_plugins(related_plugins)

            num_installed = len(install_status["installed"])
            num_failed = len(install_status["errors"])

            fg = "green"
            if num_failed >= 0 and num_installed == 0:
                fg = "red"
            elif num_failed > 0 and num_installed > 0:
                fg = "yellow"

            click.secho(
                f"Installed {num_installed}/{num_installed+num_failed} related plugins.",
                fg=fg,
            )



from itertools import groupby
from flask import request, jsonify, g

from meltano.core.error import PluginInstallError
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.plugin_install_service import PluginInstallService
from flask_security import roles_required
from meltano.api.api_blueprint import APIBlueprint


pluginsBP = APIBlueprint("plugins", __name__)


@pluginsBP.errorhandler(PluginInstallError)
def _handle(ex):
    return (jsonify({"error": True, "code": str(ex)}), 502)


@pluginsBP.route("/all", methods=["GET"])
def all():
    project = Project.find()
    discovery = PluginDiscoveryService(project)
    ordered_plugins = {}

    for type, plugins in groupby(discovery.plugins(), key=lambda p: p.type):
        canonical_plugins = []
        for plugin in plugins:
            canonical_plugin = plugin.canonical()

            # let's remove all the settings related data
            canonical_plugin.pop("settings", None)
            canonical_plugin.pop("settings_group_validation", None)

            canonical_plugins.append(canonical_plugin)

        ordered_plugins[type] = canonical_plugins

    return jsonify(ordered_plugins)


@pluginsBP.route("/installed", methods=["GET"])
def installed():
    """Returns JSON of all installed plugins

    Fuses the discovery.yml data with meltano.yml data and sorts each type alphabetically by name
    """

    project = Project.find()
    config = ConfigService(project)
    discovery = PluginDiscoveryService(project)
    installed_plugins = {}

    # merge definitions
    for plugin in sorted(config.plugins(), key=lambda x: x.name):
        try:
            definition = discovery.find_plugin(plugin.type, plugin.name)
            merged_plugin_definition = {**definition.canonical(), **plugin.canonical()}
        except PluginNotFoundError:
            merged_plugin_definition = {**plugin.canonical()}

        merged_plugin_definition.pop("settings", None)
        merged_plugin_definition.pop("select", None)

        if not plugin.type in installed_plugins:
            installed_plugins[plugin.type] = []

        installed_plugins[plugin.type].append(merged_plugin_definition)

    return jsonify({**project.meltano.canonical(), "plugins": installed_plugins})


@pluginsBP.route("/add", methods=["POST"])
@roles_required("admin")
def add():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()
    add_service = ProjectAddService(project)
    plugin = add_service.add(plugin_type, plugin_name)

    return jsonify(plugin.canonical())


@pluginsBP.route("/install/batch", methods=["POST"])
@roles_required("admin")
def install_batch():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()

    # We use the DiscoveryService rather than the ConfigService because the
    # plugin may not actually be installed yet at this point.
    discovery = PluginDiscoveryService(project)
    plugin = discovery.find_plugin(plugin_type, plugin_name)

    add_service = ProjectAddService(project)
    related_plugins = add_service.add_related(plugin)

    install_service = PluginInstallService(project)
    install_service.install_plugins(related_plugins)

    return jsonify([plugin.canonical() for plugin in related_plugins])


@pluginsBP.route("/install", methods=["POST"])
@roles_required("admin")
def install():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()

    config_service = ConfigService(project)
    plugin = config_service.find_plugin(plugin_name, plugin_type=plugin_type)

    install_service = PluginInstallService(project)
    install_service.install_plugin(plugin)

    return jsonify(plugin.canonical())



import os
import yaml
import click
import shutil
import logging
from functools import singledispatch
from typing import List, Dict
from pathlib import Path

from meltano.core.utils import truthy
import meltano.core.bundle as bundle
from meltano.core.plugin import PluginType
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.config_service import ConfigService, PluginMissingError
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationService
from .project import Project
from .venv_service import VenvService


class ProjectInitServiceError(Exception):
    pass


@singledispatch
def visit(node, executor):
    pass


@visit.register(dict)
def _(node: Dict, target_path: Path = None):
    created = []

    logging.debug(f"{target_path}")
    for name, definition in node.items():
        directory = target_path.joinpath(os.path.dirname(name))

        # always create the base directory
        os.makedirs(directory, exist_ok=True)

        # recurse for the nested definition
        created += visit(definition, target_path.joinpath(name))

    return created


@visit.register(str)
def _(node: str, target_path: Path):
    """
    Create the file using either the raw content or a bundled file.
    """
    logging.debug(f"{target_path}")
    if node.startswith("bundle://"):
        # copy from the bundle
        _, path = node.split("bundle://")
        path = bundle.find(path)

        logging.debug(f"{path} → {target_path}")
        if path.is_dir():
            shutil.copytree(path, target_path)
        else:
            shutil.copy(path, target_path)
    else:
        # write the content
        with target_path.open("w") as target:
            target.write(node)

    return [target_path]


class ProjectInitService:
    def __init__(self, project_name):
        self.initialize_file = bundle.find("initialize.yml")
        self.project_name = project_name.lower()

    def init(self, create_system_database=True, activate=True) -> Project:
        try:
            os.mkdir(self.project_name)
        except Exception as e:
            raise ProjectInitServiceError
        click.secho(f"Created", fg="blue", nl=False)
        click.echo(f" {self.project_name}")

        self.project = Project(self.project_name)
        self.create_files()

        if activate:
            Project.activate(self.project)

        if create_system_database:
            self.create_system_database()

        return self.project

    def create_files(self):
        click.secho(f"Creating project files...", fg="blue")
        default_project_yaml = yaml.safe_load(open(self.initialize_file))
        for path in visit(default_project_yaml, Path(self.project_name)):
            click.secho(f"Created", fg="blue", nl=False)
            click.echo(f" {path}")

    def create_system_database(self):
        click.secho(f"Creating system database...", fg="blue")
        # register the system database connection
        engine_uri = f"sqlite:///{self.project.root}/.meltano/meltano.db"
        engine, _ = project_engine(self.project, engine_uri, default=True)

        migration_service = MigrationService(engine)
        migration_service.upgrade()
        migration_service.seed(self.project)

    def install_plugin(self, plugin_type, plugin_name):
        try:
            self.config_service.find_plugin(plugin_name)
        except PluginMissingError:
            plugin = self.add_service.add(plugin_type, plugin_name)
            self.install_service.install_plugin(plugin)

    def echo_instructions(self):
        click.echo()
        click.secho("Project", nl=False)
        click.secho(f" {self.project_name}", fg="green", nl=False)
        click.echo(" has been created.")

        click.echo("\nNext steps:")
        click.secho("\tcd ", nl=False)
        click.secho(self.project_name, fg="green")
        click.echo("\tVisit https://meltano.com/ to learn where to go from here")

        click.echo()
        click.secho("> ", fg="bright_black", nl=False)

        click.secho("Meltano sends anonymous usage data that help improve the product.")

        click.secho("> ", fg="bright_black", nl=False)

        click.echo("You can opt-out for new, existing, or all projects.")

        click.secho("> ", fg="bright_black", nl=False)

        click.secho(
            "https://meltano.com/docs/environment-variables.html#anonymous-usage-data",
            fg="cyan",
        )

    def join_with_project_base(self, filename):
        return os.path.join(".", self.project_name, filename)



import subprocess
import requests
import logging
from functools import wraps
from urllib.parse import urlsplit
from flask import (
    Blueprint,
    render_template,
    request,
    jsonify,
    redirect,
    make_response,
    g,
    current_app,
)
from flask_login import current_user
from flask_security import roles_required, logout_user
from jinja2 import TemplateNotFound

import meltano
from flask_security import roles_required
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import is_unauthorized
from meltano.core.utils import truthy

logger = logging.getLogger(__name__)
root = Blueprint("root", __name__)


def redirect_to_login_if_auth_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if not is_unauthorized():
            return f(*args, **kwargs)

        return current_app.login_manager.unauthorized()

    return decorated


@root.errorhandler(500)
def internal_error(exception):
    logger.info(f"[{request.remote_addr}], error: {exception}")
    return jsonify({"error": str(exception)}), 500


@root.route("/-/embed/", defaults={"token": ""})
@root.route("/-/embed/<token>")
def embed(token):
    try:
        return render_template("embed.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/webapp of the Meltano project."


# this route is a catch-all route to forward
# all oustanding request (not caught by any route)
# to the front-end.
@root.route("/", defaults={"path": ""})
@root.route("/<path:path>")
@redirect_to_login_if_auth_required
def default(path):
    try:
        return render_template("webapp.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/webapp of the Meltano project."


@root.route("/upgrade", methods=["POST"])
@roles_required("admin")
def upgrade():
    meltano.api.executor.upgrade()
    return "Meltano update in progress.", 201


@root.route("/version")
def version():
    response_payload = {"version": meltano.__version__}

    if truthy(request.args.get("include_latest")):
        res = requests.get("https://pypi.org/pypi/meltano/json")
        pypi_payload = res.json()
        response_payload["latest_version"] = pypi_payload["info"]["version"]

    return jsonify(response_payload)


@root.route("/bootstrap")
@redirect_to_login_if_auth_required
def bootstrap():
    return redirect(current_app.config["MELTANO_UI_URL"])


@root.route("/echo", methods=["POST"])
def echo():
    payload = request.get_json()
    print(payload)
    return jsonify(payload)


api_root = APIBlueprint("api_root", __name__, url_prefix="/api/v1/")


@api_root.route("/identity")
def identity():
    if current_user.is_anonymous:
        return jsonify(
            {
                "username": "Anonymous",
                "anonymous": True,
                "can_sign_in": current_app.config["MELTANO_AUTHENTICATION"],
            }
        )

    return jsonify(
        {"username": current_user.username, "anonymous": False, "can_sign_in": False}
    )



import click
from . import cli
from .params import project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.config_service import ConfigService
from meltano.core.plugin import PluginType
from meltano.core.project import Project, ProjectNotFound
from meltano.core.project_add_service import ProjectAddService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine


def install_status_update(data):
    plugin = data["plugin"]

    if data["status"] == "error":
        click.secho(data["message"], fg="red")
        click.secho(data["details"], err=True)
    if data["status"] == "warning":
        click.secho(f"Warning! {data['message']}.", fg="yellow")
    if data["status"] == "success":
        click.secho(f"Installed '{plugin.name}'.", fg="green")


@cli.command()
@click.argument(
    "plugin_type", type=click.Choice([*list(PluginType), "all"]), default="all"
)
@click.option("--include-related", is_flag=True)
@project(migrate=True)
def install(project, plugin_type, include_related):
    """
    Installs all the dependencies of your project based on the meltano.yml file.
    Read more at https://www.meltano.com/docs/command-line-interface.html.
    """
    config_service = ConfigService(project)

    if include_related:
        discovery_service = PluginDiscoveryService(project)
        add_service = ProjectAddService(
            project,
            config_service=config_service,
            plugin_discovery_service=discovery_service,
        )

        installed_plugins = config_service.plugins()
        for plugin_install in installed_plugins:
            try:
                plugin_def = discovery_service.find_plugin(
                    plugin_install.type, plugin_install.name
                )
            except PluginNotFoundError:
                continue

            related_plugins = add_service.add_related(plugin_def)
            for plugin in related_plugins:
                click.secho(
                    f"Added '{plugin.name}' to your Meltano project because it is related to '{plugin_def.name}'.",
                    fg="green",
                )

    if plugin_type == "all":
        plugins = config_service.plugins()
    else:
        plugins = config_service.get_plugins_of_type(PluginType(plugin_type))

    install_service = PluginInstallService(project)
    install_status = install_service.install_plugins(
        plugins, status_cb=install_status_update
    )
    num_installed = len(install_status["installed"])
    num_failed = len(install_status["errors"])

    fg = "green"
    if num_failed >= 0 and num_installed == 0:
        fg = "red"
    elif num_failed > 0 and num_installed > 0:
        fg = "yellow"

    click.secho(f"{num_installed}/{num_installed+num_failed} plugins installed.", fg=fg)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_install()



import subprocess
import json
import logging
import asyncio
import os
import sys

from pathlib import Path
from datetime import datetime
from enum import IntFlag

from . import Runner
from meltano.core.job import Job, JobFinder
from meltano.core.plugin_invoker import invoker_factory, PluginInvoker
from meltano.core.config_service import ConfigService
from meltano.core.plugin.singer import SingerTap, SingerTarget, PluginType
from meltano.core.connection_service import ConnectionService
from meltano.core.utils import file_has_data
from meltano.core.logging import capture_subprocess_output
from meltano.core.elt_context import ELTContext


class SingerPayload(IntFlag):
    STATE = 1


class SingerRunner(Runner):
    def __init__(
        self,
        elt_context: ELTContext,
        config_service: ConfigService = None,
        connection_service: ConnectionService = None,
        **config,
    ):
        self.context = elt_context
        self.config = config
        self.config_service = config_service or ConfigService(elt_context.project)
        self.connection_service = connection_service or ConnectionService(elt_context)

        self.tap_config_dir = Path(config.get("tap_config_dir", "/etc/singer/tap"))
        self.target_config_dir = Path(
            config.get("target_config_dir", "/etc/singer/target")
        )

    def stop(self, process, **wait_args):
        while True:
            try:
                code = process.wait(**wait_args)
                logging.debug(f"{process} exited with {code}")
                return code
            except subprocess.TimeoutExpired:
                process.kill()
                logging.error(f"{process} was killed.")

    async def invoke(self, tap: PluginInvoker, target: PluginInvoker):
        try:
            # use standard pipes here because we
            # don't need async processing between the
            # tap and target.
            target_in, tap_out = os.pipe()

            p_target, p_tap = None, None
            p_target = await target.invoke_async(
                stdin=target_in,
                stdout=asyncio.subprocess.PIPE,  # state log
                stderr=asyncio.subprocess.PIPE,  # Target err for logging it
            )
            os.close(target_in)

            p_tap = await tap.invoke_async(
                stdout=tap_out, stderr=asyncio.subprocess.PIPE
            )
            os.close(tap_out)
        except Exception as err:
            if p_tap:
                p_tap.kill()
            if p_target:
                p_target.kill()
            raise Exception(f"Cannot start tap or target: {err}")

        # receive the target stdout and update the current job
        # for each line
        await asyncio.wait(
            [
                self.bookmark(p_target.stdout),
                capture_subprocess_output(p_tap.stderr, sys.stderr),
                capture_subprocess_output(p_target.stderr, sys.stderr),
                p_target.wait(),
                p_tap.wait(),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )

        # at this point, something already stopped, the other component
        # should die soon because of a SIGPIPE
        target_code = await p_target.wait()
        tap_code = await p_tap.wait()

        if any((tap_code, target_code)):
            raise Exception(
                f"Subprocesses didn't exit cleanly: tap({tap_code}), target({target_code})"
            )

    def restore_bookmark(self, session, tap: PluginInvoker):
        if self.context.job is None:
            logging.info(
                f"Running outside a Job context: incremental state could not be loaded."
            )
            return

        # the `state.json` is stored in the database
        finder = JobFinder(self.context.job.job_id)
        state_job = finder.latest_with_payload(session, flags=SingerPayload.STATE)

        if state_job and "singer_state" in state_job.payload:
            logging.info(f"Found state from {state_job.started_at}.")
            with tap.files["state"].open("w+") as state:
                json.dump(state_job.payload["singer_state"], state)
        else:
            logging.warning("No state was found, complete import.")

            # Delete state left over from different pipeline run for same extractor
            try:
                os.remove(tap.files["state"])
            except OSError:
                pass

    def bookmark_state(self, new_state: str):
        if self.context.job is None:
            logging.info(
                f"Running outside a Job context: incremental state could not be updated."
            )
            return

        try:
            new_state = json.loads(new_state)
            job = self.context.job

            job.payload["singer_state"] = new_state
            job.payload_flags |= SingerPayload.STATE
            job.ended_at = datetime.utcnow()
            logging.info(f"Incremental state has been updated at {job.ended_at}.")
            logging.debug(f"Incremental state: {new_state}")
        except Exception as err:
            logging.warning(
                "Received state is invalid, incremental state has not been updated"
            )

    async def bookmark(self, target_stream):
        while not target_stream.at_eof():
            last_state = await target_stream.readline()
            if not last_state:
                continue

            self.bookmark_state(last_state)

    def dry_run(self, tap: PluginInvoker, target: PluginInvoker):
        logging.info("Dry run:")
        logging.info(f"\textractor: {tap.plugin.name} at '{tap.exec_path()}'")
        logging.info(f"\tloader: {target.plugin.name} at '{target.exec_path()}'")

    def run(self, session, dry_run=False):
        tap = self.context.extractor_invoker()
        target = self.context.loader_invoker()

        if dry_run:
            return self.dry_run(tap, target)

        # Sets the proper `schema` for the target from the ELTContext
        target_elt_params = self.connection_service.load_params()
        target.plugin_config.update(target_elt_params)

        tap.prepare(session)
        target.prepare(session)

        self.restore_bookmark(session, tap)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.invoke(tap, target))



import os
import click
import logging
import sqlalchemy
from pathlib import Path
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.runtime.migration import MigrationContext
from alembic import command

from meltano.migrations import MIGRATION_DIR, LOCK_PATH
from meltano.core.db import project_engine
from meltano.api.models.security import Role, RolePermissions


class MigrationUneededException(Exception):
    """Occurs when no migrations are needed."""

    pass


class MigrationService:
    def __init__(self, engine):
        self.engine = engine

    def ensure_migration_needed(self, script, context, target_revision):
        current_head = context.get_current_revision()

        for rev in script.iterate_revisions(current_head, "base"):
            if rev.revision == target_revision:
                raise MigrationUneededException

    def upgrade(self):
        conn = self.engine.connect()
        cfg = Config()

        # this connection is used in `env.py` for the migrations
        cfg.attributes["connection"] = conn
        cfg.set_main_option("script_location", str(MIGRATION_DIR))
        script = ScriptDirectory.from_config(cfg)
        # let's make sure we actually need to migrate
        context = MigrationContext.configure(conn)

        try:
            # try to find the locked version
            HEAD = LOCK_PATH.open().read().strip()
            self.ensure_migration_needed(script, context, HEAD)

            click.secho(f"Upgrading database to {HEAD}")
            command.upgrade(cfg, HEAD)
        except FileNotFoundError:
            raise click.ClickException(
                "Cannot upgrade the system database, revision lock not found."
            )
        except MigrationUneededException:
            click.secho(f"System database up-to-date.")
        except Exception as err:
            click.secho(
                f"Cannot upgrade the system database. It might be corrupted or was created before database migrations where introduced (v0.34.0)",
                fg="yellow",
                err=True,
            )
            logging.exception(err)
        finally:
            conn.close()

    def seed(self, project):
        _, Session = project_engine(project)

        try:
            session = Session()

            if not session.query(Role).filter_by(name="admin").first():
                session.add(
                    Role(
                        name="admin",
                        description="Meltano Admin",
                        permissions=[
                            RolePermissions(type="view:design", context="*"),
                            RolePermissions(type="view:reports", context="*"),
                            RolePermissions(type="modify:acl", context="*"),
                        ],
                    )
                )

            if not session.query(Role).filter_by(name="regular").first():
                session.merge(Role(name="regular", description="Meltano User"))

            # add the universal permissions to Admin
            admin = session.query(Role).filter_by(name="admin").one()
            try:
                session.query(RolePermissions).filter_by(
                    role=admin, type="*", context="*"
                ).one()
            except sqlalchemy.orm.exc.NoResultFound:
                admin.permissions.append(RolePermissions(type="*", context="*"))

            session.commit()
        finally:
            session.close()



import fasteners
import logging
import os
import sys
import threading
import yaml
from copy import deepcopy
from contextlib import contextmanager
from dotenv import load_dotenv
from functools import wraps
from pathlib import Path
from typing import Union, Dict
from atomicwrites import atomic_write

from .error import Error
from .behavior.versioned import Versioned
from .utils import makedirs, slugify
from .meltano_file import MeltanoFile


class ProjectNotFound(Error):
    def __init__(self):
        super().__init__(
            f"Cannot find `{os.getcwd()}/meltano.yml`. Are you in a meltano project?"
        )


class Project(Versioned):
    """
    Represent the current Meltano project from a file-system
    perspective.
    """

    __version__ = 1
    _activate_lock = threading.Lock()
    _find_lock = threading.Lock()
    _meltano_rw_lock = fasteners.ReaderWriterLock()
    _default = None

    def __init__(self, root: Union[Path, str] = None):
        self.root = Path(root or os.getcwd()).resolve()
        self._meltano_ip_lock = fasteners.InterProcessLock(
            self.run_dir("meltano.yml.lock")
        )

    @classmethod
    @fasteners.locked(lock="_activate_lock")
    def activate(cls, project: "Project"):
        project.ensure_compatible()

        # helpful to refer to the current absolute project path
        os.environ["MELTANO_PROJECT_ROOT"] = str(project.root)

        # create a symlink to our current binary
        try:
            executable = Path(os.path.dirname(sys.executable), "meltano")
            if executable.is_file():
                project.run_dir().joinpath("bin").symlink_to(executable)
        except FileExistsError:
            pass

        load_dotenv(dotenv_path=project.root.joinpath(".env"), override=True)
        logging.debug(f"Activated project at {project.root}")

        # set the default project
        cls._default = project

    @property
    def backend_version(self):
        return self.meltano.version

    @classmethod
    @fasteners.locked(lock="_find_lock")
    def find(cls, from_dir: Union[Path, str] = None, activate=True):
        if cls._default:
            return cls._default

        project = Project(from_dir)

        if not project.meltanofile.exists():
            raise ProjectNotFound()

        # if we activate a project using `find()`, it should
        # be set as the default project for future `find()`
        if activate:
            cls.activate(project)

        return project

    @property
    def meltano(self) -> Dict:
        """Return a copy of the current meltano config"""
        # fmt: off
        with self._meltano_rw_lock.read_lock(), \
             self.meltanofile.open() as meltanofile:
            return MeltanoFile.parse(yaml.safe_load(meltanofile))
        # fmt: on

    @contextmanager
    def meltano_update(self):
        """
        Yield the current meltano configuration and update the meltanofile
        if the context ends gracefully.
        """
        # fmt: off
        with self._meltano_rw_lock.write_lock(), \
            self._meltano_ip_lock:

            with self.meltanofile.open() as meltanofile:
                # read the latest version
                meltano_update = MeltanoFile.parse(yaml.safe_load(meltanofile))

            yield meltano_update

            try:
                with atomic_write(self.meltanofile, overwrite=True) as meltanofile:
                    # update if everything is fine
                    yaml.dump(meltano_update, meltanofile, default_flow_style=False)
            except Exception as err:
                logging.critical(f"Could not update meltano.yml: {err}")
                raise
        # fmt: on

    def root_dir(self, *joinpaths):
        return self.root.joinpath(*joinpaths)

    @property
    def meltanofile(self):
        return self.root.joinpath("meltano.yml")

    @makedirs
    def meltano_dir(self, *joinpaths):
        return self.root.joinpath(".meltano", *joinpaths)

    @makedirs
    def analyze_dir(self, *joinpaths):
        return self.root_dir("analyze", *joinpaths)

    @makedirs
    def extract_dir(self, *joinpaths):
        return self.root_dir("extract", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes):
        return self.meltano_dir(*prefixes, "venv")

    @makedirs
    def run_dir(self, *joinpaths):
        return self.meltano_dir("run", *joinpaths)

    @makedirs
    def job_dir(self, job_id, *joinpaths):
        return self.run_dir("elt", slugify(job_id), *joinpaths)

    @makedirs
    def model_dir(self, *joinpaths):
        return self.meltano_dir("models", *joinpaths)

    @makedirs
    def plugin_dir(self, plugin: "PluginRef", *joinpaths):
        return self.meltano_dir(plugin.type, plugin.name, *joinpaths)

    def __eq__(self, other):
        return self.root == other.root

    def __hash__(self):
        return self.root.__hash__()



import os
import logging
import click
import datetime

from . import cli
from .params import project
from click_default_group import DefaultGroup
from meltano.core.project import Project, ProjectNotFound
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.schedule_service import ScheduleService, ScheduleAlreadyExistsError
from meltano.core.db import project_engine


@cli.group(cls=DefaultGroup, default="add")
@click.pass_context
@project(migrate=True)
def schedule(project, ctx):
    _, Session = project_engine(project)
    session = Session()

    ctx.obj["schedule_service"] = schedule_service = ScheduleService(project)
    ctx.obj["session"] = session


@schedule.command(short_help="[default] Add a new schedule")
@click.argument("name")
@click.argument("extractor")
@click.argument("loader")
@click.argument("interval")
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option("--start-date", type=click.DateTime(), default=None)
@click.pass_context
def add(ctx, name, extractor, loader, transform, interval, start_date):
    """
    Add a new schedule

    \b
    NAME:\tThe schedule name, must be unique
    EXTRACTOR:\tWhich extractor should be used
    LOADER:\tWhich loader should be used
    INTERVAL:\tCron-like syntax to specify the schedule interval (@daily, @hourly, etc…)
    """

    schedule_service = ctx.obj["schedule_service"]
    session = ctx.obj["session"]

    try:
        tracker = GoogleAnalyticsTracker(schedule_service.project)
        schedule = schedule_service.add(
            session, name, extractor, loader, transform, interval, start_date
        )

        tracker.track_meltano_schedule(schedule)
        click.echo(f"Scheduled '{schedule.name}' at {schedule.interval}")
    except ScheduleAlreadyExistsError as serr:
        click.secho(f"Schedule '{serr.schedule.name}' already exists.", fg="yellow")
    except Exception as err:
        click.secho(f"Scheduling failed: {err}", fg="red", err=True)
        raise click.Abort()


@schedule.command()
@click.pass_context
def list(ctx):
    schedule_service = ctx.obj["schedule_service"]
    transform_elt_markers = {"run": ("→", "→"), "only": ("×", "→"), "skip": ("→", "x")}

    for schedule in schedule_service.schedules():
        markers = transform_elt_markers[schedule.transform]
        click.echo(
            f"[{schedule.interval}] {schedule.name}: {schedule.extractor} {markers[0]} {schedule.loader} {markers[1]} transforms"
        )



"""
This module files should be created via
`make bundle` at the project root.
"""

import os
from pathlib import Path


def root() -> Path:
    return Path(os.path.dirname(__file__))


def find(file_path: Path) -> Path:
    return root().joinpath(file_path)



import os
import json
import yaml
import logging

from .project import Project
from .plugin import Plugin, PluginInstall


class TransformAddService:
    def __init__(self, project: Project):
        self.project = project
        self.packages_file = project.root.joinpath("transform").joinpath("packages.yml")
        self.dbt_project_file = project.root.joinpath("transform").joinpath(
            "dbt_project.yml"
        )

    def add_to_packages(self, plugin: Plugin):
        if not os.path.exists(self.packages_file):
            with open(self.packages_file, "w"):
                pass

        package_yaml = yaml.safe_load(self.packages_file.open()) or {"packages": []}

        for package in package_yaml["packages"]:
            if package.get("git", "") == plugin.pip_url:
                return

        package_yaml["packages"].append({"git": plugin.pip_url})

        with open(self.packages_file, "w") as f:
            f.write(yaml.dump(package_yaml, default_flow_style=False))

    def update_dbt_project(self, plugin: PluginInstall):
        transform_name = plugin.name.replace("-", "_")
        dbt_project_yaml = yaml.safe_load(self.dbt_project_file.open())

        dbt_project_yaml["models"][transform_name] = plugin._extras

        with open(self.dbt_project_file, "w") as f:
            f.write(yaml.dump(dbt_project_yaml, default_flow_style=False))



import yaml
import copy
from datetime import datetime, date


from typing import Iterable, List
from meltano.core.plugin import PluginInstall
from meltano.core.plugin.factory import plugin_factory
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq


VERSION = 1


class Schedule(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        extractor: str = None,
        loader: str = None,
        transform: str = None,
        interval: str = None,
        start_date: datetime = None,
        env={},
    ):
        super().__init__()

        self.name = name
        self.extractor = extractor
        self.loader = loader
        self.transform = transform
        self.interval = interval
        self.start_date = start_date
        self.env = env


class MeltanoFile(Canonical):
    def __init__(self, **attrs):
        super().__init__(
            version=int(attrs.pop("version", VERSION)),
            plugins=self.load_plugins(attrs.pop("plugins", {})),
            schedules=self.load_schedules(attrs.pop("schedules", [])),
            send_anonymous_usage_stats=attrs.pop("send_anonymous_usage_stats", None),
            project_id=attrs.pop("project_id", None),
            **attrs
        )

    def load_plugins(self, plugins) -> Canonical:
        """Parse the meltano.yml file and return it as `PluginInstall` instances."""
        plugin_type_plugins = Canonical()

        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        for plugin_type, plugin_defs in plugins.items():
            plugin_type_plugins[plugin_type] = []

            for plugin_def in plugin_defs:
                plugin = plugin_factory(plugin_type, plugin_def)
                plugin_type_plugins[plugin_type].append(plugin)

        return plugin_type_plugins

    def load_schedules(self, schedules) -> List[Schedule]:
        return list(map(Schedule.parse, schedules))



import os
import yaml
import click
import logging
from urllib.parse import urlparse

from meltano.core.project_init_service import (
    ProjectInitService,
    ProjectInitServiceError,
)
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.error import SubprocessError
from . import cli

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"


@cli.command()
@click.pass_context
@click.argument("project_name")
@click.option(
    "--no_usage_stats", help="Do not send anonymous usage stats.", is_flag=True
)
def init(ctx, project_name, no_usage_stats):
    """
    Creates a new Meltano project
    """
    if ctx.obj["project"]:
        logging.warning(f"Found meltano project at: {ctx.obj['project'].root}")
        raise click.ClickException(
            "`meltano init` cannot run inside a Meltano project."
        )

    init_service = ProjectInitService(project_name)
    try:
        project = init_service.init()
        init_service.echo_instructions()

        tracker = GoogleAnalyticsTracker(project)
        if no_usage_stats:
            tracker.update_permission_to_track(False)
        else:
            tracker.track_meltano_init(project_name=project_name)
    except ProjectInitServiceError as e:
        click.secho(f"Directory {project_name} already exists!", fg="red")
        raise click.Abort()
    except SubprocessError as proc_err:
        click.secho(str(proc_err), fg="red")
        stderr = proc_err.process.stderr
        if not isinstance(stderr, str):
            stderr = stderr.read()
        click.secho(stderr, err=True)
        raise click.Abort()



import asyncio
import logging
import os
import re
import sys


LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
CURRENT_LEVEL = os.getenv("MELTANO_LOG_LEVEL", "info")
FORMAT = "[%(asctime)s|%(levelname).1s|%(threadName)10s|%(name)s] %(message)s"


def current_log_level():
    return LEVELS[CURRENT_LEVEL]


def setup_logging(log_level=CURRENT_LEVEL):
    # setting that for any subprocess started by this process
    os.environ["MELTANO_LOG_LEVEL"] = log_level

    logging.basicConfig(stream=sys.stderr, format=FORMAT, level=LEVELS[log_level])


def remove_ansi_escape_sequences(line):
    """
    Remove ANSI escape sequences that are used for adding colors in
     terminals, so that only the text is logged in a file
    """
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", line)


async def capture_subprocess_output(stream_captured, output_stream):
    """
    Capture in real time the output stream of a suprocess that is run async.
    The stream has been set to asyncio.subprocess.PIPE and is provided using
     stream_captured to this function.
    As new lines are captured for stream_captured, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
     for the subprocess to end.
    """
    while not stream_captured.at_eof():
        line = await stream_captured.readline()
        if not line:
            continue

        output_stream.write(line.decode())



import os
import click
import subprocess
import psutil
import meltano
import logging
from pathlib import Path
from sqlalchemy import create_engine

from meltano.core.project import Project
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationService
from meltano.core.upgrade_service import UpgradeService
from . import cli
from .params import project, db_options


class UpgradeError(Exception):
    """Occurs when the Meltano upgrade fails"""

    pass


@cli.command()
@click.option("--pip_url", type=str, envvar="MELTANO_UPGRADE_PIP_URL")
@click.option("--force", is_flag=True, default=False, envvar="MELTANO_UPGRADE_FORCE")
@project()
@click.pass_context
def upgrade(ctx, project, **kwargs):
    engine, _ = project_engine(project)
    upgrade_service = UpgradeService(engine, project)

    try:
        upgrade_service.upgrade(**kwargs)
        upgrade_service.reload()
    except UpgradeError as up:
        click.secho(str(up), fg="red")
        raise click.Abort()



import os
import json
import yaml
import logging
import click

from .project import Project
from .plugin import PluginType, PluginInstall
from .plugin.factory import plugin_factory
from .config_service import ConfigService


class ProjectAddCustomService:
    def __init__(self, project: Project, config_service: ConfigService = None):
        self.project = project
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str):
        click.secho(
            f"Adding new custom {plugin_type.cli_command} plugin with name '{plugin_name}'...",
            fg="green",
        )
        click.echo()

        click.echo(
            f"Specify the plugin's {click.style('namespace', fg='blue')}, which will serve as the:"
        )
        click.echo("- prefix for configuration environment variables")
        click.echo("- identifier to find related/compatible plugins")
        click.echo("- target database schema when used with")
        click.echo("  loader target-postgres or target-snowflake")
        click.echo()
        click.echo(
            "Hit Return to accept the default: plugin name with underscores instead of dashes"
        )
        click.echo()

        default_namespace = plugin_name.replace("-", "_")
        namespace = click.prompt(
            click.style("(namespace)", fg="blue"), type=str, default=default_namespace
        )

        click.echo()
        click.echo(
            f"Specify the plugin's {click.style('`pip install` argument', fg='blue')}, for example:"
        )
        click.echo("- PyPI package name:")
        click.echo(f"\t{plugin_name}")
        click.echo("- VCS repository URL:")
        click.echo(f"\tgit+https://gitlab.com/meltano/{plugin_name}.git")
        click.echo("- local directory, in editable/development mode:")
        click.echo(f"\t-e extract/{plugin_name}")
        click.echo()
        click.echo("Default: plugin name as PyPI package name")
        click.echo()

        pip_url = click.prompt(
            click.style("(pip_url)", fg="blue"), type=str, default=plugin_name
        )

        click.echo()
        click.echo(f"Specify the package's {click.style('executable name', fg='blue')}")
        click.echo()
        click.echo("Default: package name derived from `pip_url`")
        click.echo()

        package_name, _ = os.path.splitext(os.path.basename(pip_url))
        executable = click.prompt(
            click.style("(executable)", fg="blue"), default=package_name
        )

        capabilities = []
        if plugin_type == PluginType.EXTRACTORS:
            click.echo()
            click.echo(
                f"Specify the tap's {click.style('supported Singer features', fg='blue')} (executable flags), for example:"
            )
            click.echo("\t`catalog`: supports the `--catalog` flag")
            click.echo("\t`discover`: supports the `--discover` flag")
            click.echo("\t`properties`: supports the `--properties` flag")
            click.echo("\t`state`: supports the `--state` flag")
            click.echo()
            click.echo(
                "To find out what features a tap supports, reference its documentation or try one"
            )
            click.echo(
                "of the tricks under https://meltano.com/docs/contributor-guide.html#how-to-test-a-tap."
            )
            click.echo()
            click.echo("Multiple capabilities can be separated using commas.")
            click.echo()
            click.echo("Default: no capabilities")
            click.echo()

            capabilities = click.prompt(
                click.style("(capabilities)", fg="blue"),
                type=list,
                default=[],
                value_proc=lambda value: [c.strip() for c in value.split(",")],
            )

        settings = []
        if plugin_type in (PluginType.EXTRACTORS, PluginType.LOADERS):
            singer_type = "tap" if plugin_type == PluginType.EXTRACTORS else "target"

            click.echo()
            click.echo(
                f"Specify the {singer_type}'s {click.style('supported settings', fg='blue')} (`config.json` keys)"
            )
            click.echo()
            click.echo("Nested properties can be represented using the `.` seperator,")
            click.echo('e.g. `auth.username` for `{ "auth": { "username": value } }`.')
            click.echo()
            click.echo(
                f"To find out what settings a {singer_type} supports, reference its documentation."
            )
            click.echo()
            click.echo("Multiple setting names (keys) can be separated using commas.")
            click.echo()
            click.echo("Default: no settings")
            click.echo()

            settings = click.prompt(
                click.style("(settings)", fg="blue"),
                type=list,
                default=[],
                value_proc=lambda value: [c.strip() for c in value.split(",")],
            )

        # manually create the generic PluginInstall to save it
        # as a custom plugin
        install = PluginInstall(
            plugin_type,
            plugin_name,
            pip_url,
            capabilities=capabilities,
            namespace=namespace,
            executable=executable,
            settings=[{"name": name} for name in settings],
        )

        installed = self.config_service.add_to_file(install)
        click.secho(
            "The plugin has been added to your `meltano.yml`.\n"
            "If your plugin requires configuration options, you can add them directly "
            "in the created plugin definition under the `settings` or `config` section.",
            fg="yellow",
        )

        return installed



import logging

from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.behavior.hookable import hook
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.venv_service import VirtualEnv
from meltano.core.m5o.reports_service import ReportAlreadyExistsError, ReportsService
from meltano.core.m5o.dashboards_service import (
    DashboardAlreadyExistsError,
    DashboardsService,
)


class RecordImporter:
    def __init__(self, packages_dir, project):
        self.packages_dir = packages_dir
        self.project = project

    @property
    def record_type(self):
        return self.__class__.RECORD_TYPE

    def import_records(self):
        logging.debug(f"Importing {self.record_type}s...")

        records = self.records_to_import()

        logging.debug(f"Found {len(records)} {self.record_type}s to import")

        for record in records:
            self.import_record(record)

        logging.debug(f"Successfully imported {len(records)} {self.record_type}s")

    def records_to_import(self):
        return self.get_package_records(f"{self.record_type}s")

    def import_record(self, importable_record):
        logging.debug(
            f"Importing {self.record_type} with name '{importable_record['name']}', ID '{importable_record['id']}'"
        )

        existing_record = self._get_record(importable_record["id"])
        if existing_record:
            logging.debug(
                f"Found existing {self.record_type} with ID '{existing_record['id']}'"
            )

            record = existing_record
            record = self.update_record_data(record, importable_record)

            save_func = self._update_record
        else:
            record = importable_record
            save_func = self._save_record

        record = self.save_record(record, save_func)

        logging.debug(
            f"Successfully imported {self.record_type} with ID '{record['id']}', name '{record['name']}'"
        )

    def update_record_data(self, record, importable_record):
        snapshots = self.snapshots_for_record(importable_record["id"])
        if not snapshots:
            logging.debug(
                f"No snapshots found: not overwriting anything with importable values"
            )
            return record

        for attribute in self.__class__.UPDATABLE_ATTRIBUTES:
            existing_value = record.get(attribute, None)
            importable_value = importable_record.get(attribute, None)

            if not importable_value:
                continue

            if importable_value == existing_value:
                logging.debug(
                    f"Existing '{attribute}' value matches importable value: nothing to do"
                )
                continue

            snapshot_values = [s.get(attribute, None) for s in snapshots]

            if existing_value in snapshot_values:
                logging.debug(
                    f"Existing '{attribute}' value matches a snapshot: overwriting with importable value"
                )
                record[attribute] = importable_value
            else:
                logging.debug(
                    f"Existing '{attribute}' value does not match any snapshots: not overwriting with importable value"
                )

        return record

    def snapshots_for_record(self, id):
        return [s for s in self.get_package_records("snapshots", id) if s["id"] == id]

    def save_record(self, record, save_func):
        original_name = record["name"]
        counter = 1

        while True:
            try:
                record = save_func(record)
                break
            except self.__class__.ALREADY_EXISTS_ERROR as e:
                if e.field == "slug":
                    counter += 1
                    record["name"] = f"{original_name} {counter}"

                    # Retry!
                    continue
                else:
                    record = e.record
                    break

        return record

    def get_package_records(self, *args):
        records_parser = M5oCollectionParser(
            self.packages_dir.joinpath(*args), self.record_type
        )
        return records_parser.parse()

    def _get_record(self, id):
        raise NotImplementedError

    def _save_record(self, record):
        raise NotImplementedError

    def _update_record(self, record):
        raise NotImplementedError


class ReportImporter(RecordImporter):
    RECORD_TYPE = M5oCollectionParserTypes.Report
    ALREADY_EXISTS_ERROR = ReportAlreadyExistsError
    UPDATABLE_ATTRIBUTES = [
        "chart_type",
        "design",
        "model",
        "namespace",
        "name",
        "query_payload",
    ]

    def __init__(self, *args):
        super().__init__(*args)

        self.reports_service = ReportsService(self.project)

    def _get_record(self, id):
        return self.reports_service.get_report(id)

    def _save_record(self, report):
        return self.reports_service.save_report(report)

    def _update_record(self, report):
        return self.reports_service.update_report(report)


class DashboardImporter(RecordImporter):
    RECORD_TYPE = M5oCollectionParserTypes.Dashboard
    ALREADY_EXISTS_ERROR = DashboardAlreadyExistsError
    UPDATABLE_ATTRIBUTES = ["name", "description", "report_ids"]

    def __init__(self, *args):
        super().__init__(*args)

        self.dashboards_service = DashboardsService(self.project)

    def import_record(self, dashboard):
        original_report_ids = dashboard["report_ids"]

        super().import_record(dashboard)

        self.add_reports_to_dashboard(dashboard, original_report_ids)

    def add_reports_to_dashboard(self, dashboard, report_ids):
        for report_id in report_ids:
            self.dashboards_service.add_report_to_dashboard(
                {"dashboard_id": dashboard["id"], "report_id": report_id}
            )

            logging.debug(
                f"Added report with ID '{report_id}' to dashboard with ID '{dashboard['id']}'"
            )

    def _get_record(self, id):
        return self.dashboards_service.get_dashboard(id)

    def _save_record(self, dashboard):
        return self.dashboards_service.save_dashboard(dashboard)

    def _update_record(self, dashboard):
        return self.dashboards_service.update_dashboard(
            {"dashboard": {"id": dashboard["id"]}, "new_settings": dashboard}
        )


class DashboardPlugin(PluginInstall):
    __plugin_type__ = PluginType.DASHBOARDS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @hook("after_install")
    def after_install(self, project, args=[]):
        venv = VirtualEnv(project.plugin_dir(self, "venv"))
        packages_dir = venv.site_packages_dir

        ReportImporter(packages_dir, project).import_records()
        DashboardImporter(packages_dir, project).import_records()



import logging
import os
import secrets

from flask import Flask, url_for, render_template
from werkzeug.contrib.fixers import ProxyFix


app = Flask(__name__, instance_path=os.getcwd(), instance_relative_config=True)

app.config.from_object("meltano.oauth.config")
app.config.from_pyfile("ui.cfg", silent=True)


@app.errorhandler(Exception)
def _handle(e):
    logging.exception(e)
    return error()


@app.route("/")
def root():
    return render_template(
        "home.html",
        providers=(
            {
                "label": "Facebook",
                "url": url_for("OAuth.Facebook.login"),
                "logo": url_for("static", filename="logos/facebook-logo.png"),
            },
            {
                "label": "Google Ads",
                "url": url_for("OAuth.GoogleAdwords.login"),
                "logo": url_for("static", filename="logos/adwords-logo.png"),
            },
        ),
    )


@app.route("/sample")
def sample():
    return render_template("token.html", token=secrets.token_hex(128))


@app.route("/error")
def error():
    return render_template("error.html")


from .providers import OAuth, facebook, google_adwords

facebook(app)
google_adwords(app)
OAuth.init_app(app)

# enable the parsing of X-Forwarded-* headers
app.wsgi_app = ProxyFix(app.wsgi_app)



import json
import logging
import subprocess
from typing import Dict
from jsonschema import Draft4Validator

from meltano.core.utils import file_has_data
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError

from . import SingerPlugin, PluginType
from .catalog import SelectExecutor


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    def exec_args(self, plugin_invoker):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            if "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]

        state_path = plugin_invoker.files["state"]
        if "state" in plugin_invoker.capabilities and file_has_data(state_path):
            args += ["--state", state_path]

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke", can_fail=True)
    def run_discovery_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args:
            return

        try:
            self.run_discovery(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            return

    def run_discovery(self, plugin_invoker, exec_args=[]):
        if not "discover" in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support schema discovery"
            )

        properties_file = plugin_invoker.files["catalog"]

        with properties_file.open("w") as catalog:
            result = plugin_invoker.invoke(
                "--discover",
                stdout=catalog,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            stdout, stderr = result.communicate()
            exit_code = result.returncode

        if exit_code != 0:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Schema discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code}: {stderr.rstrip()}"
            )

        # test for the schema to be a valid catalog
        try:
            with properties_file.open("r") as catalog:
                schema_valid = Draft4Validator.check_schema(json.load(catalog))
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                "Schema discovery failed: invalid catalog output by --discovery."
            ) from err

    @hook("before_invoke", can_fail=True)
    def apply_select_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args:
            return

        try:
            self.apply_select(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            return

    def apply_select(self, plugin_invoker, exec_args=[]):
        if (
            not "catalog" in plugin_invoker.capabilities
            and not "properties" in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support selection"
            )

        properties_file = plugin_invoker.files["catalog"]

        try:
            with properties_file.open() as catalog:
                schema = json.load(catalog)

            reset_executor = SelectExecutor(["!*.*"])
            select_executor = SelectExecutor(plugin_invoker.select)

            reset_executor.visit(schema)
            select_executor.visit(schema)

            with properties_file.open("w") as catalog:
                json.dump(schema, catalog)
        except FileNotFoundError as err:
            raise PluginExecutionError(
                f"Selection failed: catalog file is missing."
            ) from err
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Selection failed: catalog file is invalid: {properties_file}"
            ) from err



from typing import Optional
from collections import namedtuple

from meltano.core.project import Project
from meltano.core.job import Job
from meltano.core.config_service import ConfigService
from meltano.core.plugin import Plugin, PluginType, PluginRef
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_invoker import invoker_factory


class PluginContext(namedtuple("PluginContext", "ref install definition config")):
    @property
    def namespace(self):
        return self.definition.namespace

    @property
    def name(self):
        return self.install.name


class ELTContext:
    def __init__(
        self,
        project,
        loader: PluginContext,
        job: Optional[Job] = None,
        extractor: Optional[PluginContext] = None,
        plugin_settings_service: PluginSettingsService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.job = job
        self.loader = loader
        self.extractor = extractor
        self.plugin_settings_service = plugin_settings_service
        self.plugin_discovery_service = (
            plugin_discovery_service or PluginDiscoveryService(project)
        )

    @property
    def elt_run_dir(self):
        if self.job:
            self.project.job_dir(self.job.job_id, str(self.job.run_id))

    def extractor_invoker(self):
        return invoker_factory(
            self.project,
            self.extractor.install,
            run_dir=self.elt_run_dir,
            plugin_config=self.extractor.config,
            plugin_settings_service=self.plugin_settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )

    def loader_invoker(self):
        return invoker_factory(
            self.project,
            self.loader.install,
            run_dir=self.elt_run_dir,
            plugin_config=self.loader.config,
            plugin_settings_service=self.plugin_settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )


class ELTContextBuilder:
    def __init__(
        self,
        project: Project,
        config_service: ConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.plugin_discovery_service = (
            plugin_discovery_service
            or PluginDiscoveryService(project, config_service=config_service)
        )
        self.plugin_settings_service = plugin_settings_service or PluginSettingsService(
            project,
            config_service=config_service,
            plugin_discovery_service=plugin_discovery_service,
        )
        self._extractor = None
        self._loader = None
        self._job = None

    def with_extractor(self, extractor_name: str):
        self._extractor = PluginRef(PluginType.EXTRACTORS, extractor_name)

        return self

    def with_loader(self, loader_name: str):
        self._loader = PluginRef(PluginType.LOADERS, loader_name)

        return self

    def with_job(self, job: Job):
        self._job = job

        return self

    def plugin_context(self, session, plugin: PluginRef):
        return PluginContext(
            ref=plugin,
            install=self.config_service.find_plugin(
                plugin_name=plugin.name, plugin_type=plugin.type
            ),
            definition=self.plugin_discovery_service.find_plugin(
                plugin_name=plugin.name, plugin_type=plugin.type
            ),
            config=self.plugin_settings_service.as_config(session, plugin),
        )

    def context(self, session) -> ELTContext:
        return ELTContext(
            self.project,
            self.plugin_context(session, self._loader),
            job=self._job,
            extractor=self.plugin_context(session, self._extractor)
            if self._extractor
            else None,
            plugin_settings_service=self.plugin_settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )



import psutil
import logging
from pathlib import Path
from typing import Union


logger = logging.getLogger(__name__)


class PIDFile:
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)
        self._pid = None

    @property
    def pid(self):
        return self._pid or self.load_pid()

    def load_pid(self) -> int:
        try:
            with self.path.open("r") as raw:
                self._pid = int(raw.read())
        except ValueError:
            self.path.unlink()
        except FileNotFoundError:
            pass

        logger.debug(f"Loaded PID {self._pid} from {self.path}")
        return self._pid

    def write_pid(self, pid: int):
        with self.path.open("w") as raw:
            raw.write(str(pid))
            self._pid = pid

    @property
    def process(self) -> psutil.Process:
        if self.pid is None:
            raise UnknownProcessError(self)

        try:
            return psutil.Process(self.pid)
        except psutil.NoSuchProcess:
            raise UnknownProcessError(self)

    def unlink(self):
        return self.path.unlink()

    def __str__(self) -> str:
        return str(self.path)


class UnknownProcessError(Exception):
    """Occurs when the PIDFile doesn't yield a readable PID."""

    def __init__(self, pid_file: PIDFile):
        self.pid_file = pid_file

        super().__init__(f"{pid_file} doesn't refer to a valid Process.")



import logging

import fasteners
import threading

from meltano.core.project import Project
from meltano.core.m5o.m5o_file_parser import (
    MeltanoAnalysisFileParser,
    MeltanoAnalysisFileParserError,
)


class ProjectCompiler:
    """
    Meltano projects have multiple .m5o files that need to be
    compiled.

    This component is responsible for the compilation of
    the whole Meltano project.
    """

    def __init__(self, project):
        self.project = project
        self._parsed = False
        self._lock = threading.Lock()

    @property
    def source_dir(self):
        return self.project.model_dir()

    def parse(self):
        self.parser = MeltanoAnalysisFileParser(self.project)
        try:
            self.topics = self.parser.parse()
            self.package_topics = self.parser.parse_packages()
            self._parsed = True
        except MeltanoAnalysisFileParserError as e:
            logging.warning(f"Failed to compile topics: {e}")
            raise e

    @fasteners.locked
    def compile(self):
        try:
            self.parse()

            all_topics = self.topics + self.package_topics
            if all_topics:
                self.parser.compile(all_topics)

            logging.debug(f"Successfully compiled topics")
        except MeltanoAnalysisFileParserError as e:
            logging.warning(f"Failed to compile topics: {e}")
            raise e



import datetime
import subprocess
import logging
import os
from functools import partial
from flask_executor import Executor

from meltano.api.signals import PipelineSignals
from meltano.api.models import db
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.project import Project


executor = Executor()
logger = logging.getLogger(__name__)


def setup_executor(app, project):
    executor.init_app(app)


def defer_run_elt(schedule_payload: dict):
    job_id = schedule_payload["name"]
    extractor = schedule_payload["extractor"]
    loader = schedule_payload["loader"]
    transform = schedule_payload.get("transform")

    cmd = [
        "meltano",
        "elt",
        "--job_id",
        job_id,
        extractor,
        loader,
        "--transform",
        transform,
    ]

    result = subprocess.run(
        cmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env={**os.environ, "MELTANO_JOB_TRIGGER": "ui"},
    )

    # It would probably be better that we would use sqlalchemy ORM events
    # on the `Job` instance, but it would require more changes as it is
    # a `meltano.core` model.
    #
    # The benefit is that we are currently in an executor Thread, so we can
    # safely send emails without thinking about blocking any requests.
    #
    # The caveat here is that pipeline that runs from Airflow won't trigger
    # this signal, and thus won't send any notification.
    PipelineSignals.on_complete(schedule_payload, success=result.returncode == 0)


def run_elt(project: Project, schedule_payload: dict):
    job_id = schedule_payload["name"]
    executor.submit(defer_run_elt, schedule_payload)
    logger.debug(f"Defered `run_elt` to the executor with {schedule_payload}.")

    return job_id


def upgrade():
    cmd = ["meltano", "upgrade"]
    executor.submit(
        subprocess.run, cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )



from flask.signals import Namespace


class PipelineSignals:
    __namespace__ = Namespace()
    completed = __namespace__.signal("completed")

    @classmethod
    def on_complete(cls, pipeline, success: bool):
        cls.completed.send(pipeline, success=success)



import datetime
import glob
import logging
import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Union, Optional

from meltano.core.project import Project
from meltano.core.utils import slugify, makedirs


MAX_FILE_SIZE = 2_097_152  # 2MB max


class MissingJobLogException(Exception):
    """Occurs when JobLoggingService can not find a requested log."""

    pass


class SizeThresholdJobLogException(Exception):
    """Occurs when a Job log exceeds the MAX_FILE_SIZE."""

    pass


class JobLoggingService:
    def __init__(self, project: Project):
        self.project = project

    @makedirs
    def elt_dir(self, job_id, run_id, *joinpaths):
        return self.project.job_dir(job_id, str(run_id), *joinpaths)

    @contextmanager
    def create_log(self, job_id, run_id, file_name="elt.log"):
        """
        Open a new log file for logging and yield it.

        Log will be created inside the elt_dir, which is .meltano/run/elt/:job_id/:run_id
        """
        log_file_name = self.elt_dir(job_id, run_id, file_name)

        try:
            log_file = open(log_file_name, "w")
        except (OSError, IOError) as err:
            # Don't stop the Job running if you can not open the log file
            # for writting: just return /dev/null
            logging.error(
                f"Could open log file {log_file_name} for writting. Using /dev/null"
            )
            log_file = open(os.devnull, "w")

        try:
            yield log_file
        finally:
            log_file.close()

    def get_latest_log(self, job_id):
        """
        Get the contents of the most recent log for any ELT job
         that ran with the provided job_id
        """
        try:
            latest_log = next(iter(self.get_all_logs(job_id)))

            if latest_log.stat().st_size > MAX_FILE_SIZE:
                raise SizeThresholdJobLogException(
                    f"The log file size exceeds '{MAX_FILE_SIZE}'"
                )

            with latest_log.open() as f:
                return f.read()
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with id '{job_id}'"
            )
        except FileNotFoundError:
            raise MissingJobLogException(
                f"Cannot log for job with id '{job_id}': '{latest_log}' is missing."
            )

    def get_downloadable_log(self, job_id):
        """
        Get the `*.log` file of the most recent log for any ELT job
         that ran with the provided job_id
        """
        try:
            latest_log = next(iter(self.get_all_logs(job_id)))
            return str(latest_log.resolve())
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with id '{job_id}'"
            )
        except FileNotFoundError:
            raise MissingJobLogException(
                f"Cannot log for job with id '{job_id}': '{latest_log}' is missing."
            )

    def get_all_logs(self, job_id):
        """
        Get all the log files for any ELT job that ran with the provided job_id

        The result is ordered so that the most recent is first on the list
        """
        log_files = list(self.project.job_dir(job_id).glob("**/*.log"))
        log_files.sort(key=lambda path: os.stat(path).st_ctime_ns, reverse=True)

        return log_files

    def delete_all_logs(self, job_id):
        """
        Delete all the log files for any ELT job that ran with the provided job_id
        """

        try:
            shutil.rmtree(self.project.job_dir(job_id))
        except OSError:
            # If there already are no log files for this job_id, we're done.
            return



from meltano.core.project import Project
from meltano.core.schedule_service import ScheduleService

from .sql_helper import SqlHelper


class ReportsHelper:
    VERSION = "1.0.0"

    def get_report_with_query_results(self, report, today=None):
        project = Project.find()
        schedule_service = ScheduleService(project)
        sql_helper = SqlHelper()
        report_with_query_results = self.update_report_with_query_results(
            report, schedule_service, sql_helper, today=today
        )

        return report_with_query_results

    def update_report_with_query_results(
        self, report, schedule_service, sql_helper, today=None
    ):
        m5oc = sql_helper.get_m5oc_topic(report["namespace"], report["model"])
        design_helper = m5oc.design(report["design"])
        schedule = schedule_service.find_namespace_schedule(
            m5oc.content["plugin_namespace"]
        )

        report["full_design"] = design_helper.design

        query_payload = report["query_payload"]
        if today:
            query_payload["today"] = today

        sql_dict = sql_helper.get_sql(design_helper, query_payload)
        outgoing_sql = sql_dict["sql"]
        aggregates = sql_dict["aggregates"]

        report["query_results"] = sql_helper.get_query_results(
            schedule.loader, outgoing_sql
        )
        report["query_result_aggregates"] = aggregates

        return report



from flask import jsonify, request
from flask_restful import Api, Resource, fields, marshal, marshal_with
from flask_security import roles_required
from flask_principal import Permission, Need
from werkzeug.exceptions import Forbidden
from sqlalchemy.orm import joinedload
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security import users
from flask_security import roles_required
from meltano.api.models.security import db, User, Role, RolesUsers, RolePermissions
from .settings_helper import SettingsHelper


settingsBP = APIBlueprint("settings", __name__)
settingsApi = Api(settingsBP)


@settingsBP.route("/", methods=["GET"])
@roles_required("admin")
def index():
    settings_helper = SettingsHelper()
    return jsonify(settings_helper.get_connections())


@settingsBP.route("/save", methods=["POST"])
@roles_required("admin")
@roles_required("admin")
def save():
    settings_helper = SettingsHelper()
    connection = request.get_json()
    settings = settings_helper.save_connection(connection)
    return jsonify(settings)


@settingsBP.route("/delete", methods=["POST"])
@roles_required("admin")
def delete():
    settings_helper = SettingsHelper()
    connection = request.get_json()
    settings = settings_helper.delete_connection(connection)
    return jsonify(settings)


@settingsBP.route("/connections/<name>/test")
def test(name):
    current_settings = Settings.query.first().settings
    connections = current_settings["connections"]
    try:
        found_connection = next(
            connection for connection in connections if connection["name"] == name
        )
    # this is a really broad exception catch, this will swallow sneaky errors
    except Exception as e:
        found_connection = {}

    return jsonify(found_connection)


class Canonical(fields.Raw):
    def __init__(self, scopes=[], **kwargs):
        super().__init__(**kwargs)

        self._scopes = scopes

    def format(self, value):
        try:
            return value.canonical(self._scopes)
        except AttributeError:
            return value.id


class AclResource(Resource):
    UserDefinition = {
        "username": fields.String,
        "active": fields.Boolean,
        "confirmed_at": fields.DateTime,
        "roles": fields.List(Canonical),
    }

    RoleDefinition = {
        "name": fields.String,
        "description": fields.String,
        "permissions": fields.List(Canonical(scopes=[Role])),
    }

    PermissionDefinition = {"type": fields.String, "name": fields.String}

    AclDefinition = {
        "users": fields.List(fields.Nested(UserDefinition)),
        "roles": fields.List(fields.Nested(RoleDefinition)),
        "permissions": fields.List(fields.Nested(PermissionDefinition)),
    }

    @marshal_with(AclDefinition)
    @roles_required("admin")
    def get(self):
        return {
            "users": User.query.options(joinedload("roles")).all(),
            "roles": Role.query.options(joinedload("permissions")).all(),
            "permissions": [],
        }


class RolesResource(Resource):
    @roles_required("admin")
    @roles_required("admin")
    @marshal_with(AclResource.RoleDefinition)
    def post(self):
        payload = request.get_json()
        role = payload["role"]
        user = payload.get("user")

        try:
            role_name = role.pop("name")
            role = users.find_or_create_role(role_name, **role)

            if user:
                user = users.get_user(user)
                users.add_role_to_user(user, role_name)
        except KeyError:
            return "role.name must be set.", 400

        db.session.commit()
        return role, 201

    @roles_required("admin")
    @roles_required("admin")
    @marshal_with(AclResource.RoleDefinition)
    def delete(self):
        payload = request.get_json()
        role = payload["role"]
        user = payload.get("user")

        try:
            role = Role.query.filter_by(name=role["name"]).one()

            if not user:
                if role == "admin":
                    return "The `admin` role cannot be deleted.", 403

                # delete the role
                db.session.delete(role)
            else:
                # unassign the user-role
                assigned_user = role.users.filter(User.username == user).one()
                role.users.remove(assigned_user)

            db.session.commit()
        except Exception as err:
            return str(err), 400

        return role, 201


class RolePermissionsResource(Resource):
    def _parse_request(self):
        payload = request.get_json()
        return payload["role"], payload["permission_type"], payload["context"]

    @roles_required("admin")
    @roles_required("admin")
    @marshal_with(AclResource.RoleDefinition)
    def post(self):
        role, permission_type, context = self._parse_request()

        try:
            role = (
                Role.query.options(joinedload("permissions"))
                .filter_by(name=role["name"])
                .one()
            )
            perm = RolePermissions.query.filter_by(
                role=role, type=permission_type, context=context
            ).first()

            if not perm:
                perm = RolePermissions(role=role, type=permission_type, context=context)
                role.permissions.append(perm)

            db.session.commit()
        except Exception as err:
            return str(err), 400

        return role, 200

    @roles_required("admin")
    @roles_required("admin")
    @marshal_with(AclResource.RoleDefinition)
    def delete(self):
        role, permission_type, context = self._parse_request()
        role = Role.query.filter_by(name=role["name"]).one()

        try:
            perm = RolePermissions.query.filter_by(
                role=role, type=permission_type, context=context
            ).one()
            role.permissions.remove(perm)
            db.session.commit()
        except:
            return role, 200

        return role, 201


settingsApi.add_resource(AclResource, "/acl")
settingsApi.add_resource(
    RolePermissionsResource, "/acl/roles/permissions", endpoint="role_permissions"
)
settingsApi.add_resource(RolesResource, "/acl/roles", endpoint="roles")



import re

from werkzeug.utils import secure_filename

from .errors import InvalidFileNameError


def enforce_secure_filename(string: str):
    name = secure_filename(string)
    if name is "":
        raise InvalidFileNameError(name)
    return name



import click
from . import cli
from .params import project

from meltano.core.db import project_engine
from meltano.core.project import Project
from meltano.core.config_service import ConfigService
from meltano.core.plugin.settings_service import PluginSettingsService


@cli.group(invoke_without_command=True)
@click.argument("plugin_name")
@click.option("--format", default="json")
@project(migrate=True)
@click.pass_context
def config(ctx, project, plugin_name, format):
    config = ConfigService(project)
    plugin = config.find_plugin(plugin_name)

    _, Session = project_engine(project)
    session = Session()
    settings = PluginSettingsService(project)

    ctx.obj["settings"] = settings
    ctx.obj["plugin"] = plugin
    ctx.obj["session"] = session

    if ctx.invoked_subcommand is None:
        if format == "json":
            print(settings.as_config(session, plugin))

        if format == "env":
            for env, value in settings.as_env(session, plugin).items():
                print(f"{env}={value}")


@config.command()
@click.argument("setting_name")
@click.argument("value")
@click.pass_context
def set(ctx, setting_name, value):
    settings = ctx.obj["settings"]
    plugin = ctx.obj["plugin"]
    session = ctx.obj["session"]

    settings.set(session, plugin, setting_name, value)


@config.command()
@click.argument("setting_name")
@click.pass_context
def unset(ctx, setting_name):
    settings = ctx.obj["settings"]
    plugin = ctx.obj["plugin"]
    session = ctx.obj["session"]

    settings.unset(session, plugin, setting_name)


@config.command()
@click.pass_context
def reset(ctx):
    settings = ctx.obj["settings"]
    plugin = ctx.obj["plugin"]
    session = ctx.obj["session"]

    for setting in settings.definitions(plugin):
        settings.unset(session, plugin, setting.name)


@config.command()
@click.pass_context
def list(ctx):
    settings = ctx.obj["settings"]
    plugin = ctx.obj["plugin"]
    plugin_def = settings.get_definition(plugin)

    for setting_def in settings.definitions(plugin):
        env_key = settings.setting_env(setting_def, plugin_def)
        description_marker = (
            f": {setting_def.description}" if setting_def.description else ""
        )
        click.secho(f"{setting_def.name} [{env_key}]{description_marker}")



import cProfile
import pstats
import time
import logging
from flask import g
from sqlalchemy import event
from sqlalchemy.engine import Engine


logger = logging.getLogger("meltano.api.sql")
logger.setLevel(logging.DEBUG)


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    conn.info.setdefault("query_start_time", []).append(time.time())
    logger.debug("Start Query: %s", statement)


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    total = time.time() - conn.info["query_start_time"].pop(-1)
    logger.debug("Query Complete!")
    logger.debug("Total Time: %f", total)


def begin_request_profile():
    g.pr = cProfile.Profile()
    g.pr.enable()


def print_request_profile(res):
    g.pr.disable()
    g.pr.create_stats()

    stats = pstats.Stats(g.pr)
    stats.sort_stats("cumulative").print_stats(0.1)

    return res


def init(app):
    app.before_request(begin_request_profile)
    app.after_request(print_request_profile)



import logging
from flask import Blueprint, request, url_for, redirect, render_template, jsonify
from authlib.flask.client import OAuth as OAuthClient


OAuth = OAuthClient()


def facebook(app):
    OAuth.register(
        "facebook",
        access_token_url="https://graph.facebook.com/v5.0/oauth/access_token",
        client_kwargs={"scope": "ads_read ads_management manage_pages"},
        authorize_url="https://www.facebook.com/v5.0/dialog/oauth",
    )

    oauthBP = Blueprint("OAuth.Facebook", __name__, url_prefix="/facebook")

    @oauthBP.route("/", strict_slashes=False)
    def login():
        redirect_uri = url_for(".authorize", _external=True)

        return OAuth.facebook.authorize_redirect(
            redirect_uri, display="popup", auth_type="rerequest"
        )

    @oauthBP.route("/authorize")
    def authorize():
        token = OAuth.facebook.authorize_access_token()

        return render_template("token.html", token=token["access_token"])

    app.register_blueprint(oauthBP)


def google_adwords(app):
    OAuth.register(
        "google_adwords",
        authorize_url="https://accounts.google.com/o/oauth2/auth",
        client_kwargs={"scope": "https://www.googleapis.com/auth/adwords"},
        access_token_url="https://oauth2.googleapis.com/token",
    )

    oauthBP = Blueprint("OAuth.GoogleAdwords", __name__, url_prefix="/google-adwords")

    @oauthBP.route("/", strict_slashes=False)
    def login():
        redirect_uri = url_for(".authorize", _external=True)
        # `consent` prompt is required to have a refresh_token
        # `offline` access_type is required to have a refresh_token
        redirect = OAuth.google_adwords.authorize_redirect(
            redirect_uri, prompt="consent", access_type="offline"
        )
        return redirect

    @oauthBP.route("/authorize")
    def authorize():
        token = OAuth.google_adwords.authorize_access_token()

        return render_template("token.html", token=token["refresh_token"])

    app.register_blueprint(oauthBP)



import psycopg2
import os
import contextlib
import sqlalchemy.pool as pool
import logging
import weakref

from sqlalchemy import create_engine, MetaData, event
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.engine import Engine
from psycopg2.sql import Identifier, SQL


SystemMetadata = MetaData()
SystemModel = declarative_base(metadata=SystemMetadata)

# Keep a Project → Engine mapping to serve
# the same engine for the same Project
_engines = dict()


def project_engine(project, engine_uri=None, default=False) -> ("Engine", sessionmaker):
    """Creates and register a SQLAlchemy engine for a Meltano project instance."""

    # return the default engine if it is registered
    if not engine_uri and project in _engines:
        return _engines[project]
    elif (project, engine_uri) in _engines:
        return _engines[(project, engine_uri)]

    if not engine_uri:
        logging.debug(f"Can't find engine for {project}@{engine_uri}")
        raise ValueError("No engine registered for this project.")

    logging.debug(f"Creating engine {project}@{engine_uri}")
    engine = create_engine(engine_uri)

    init_hook(engine)

    create_session = sessionmaker(bind=engine)
    engine_session = (engine, create_session)

    if default:
        # register the default engine
        _engines[project] = engine_session
    else:
        _engines[(project, engine_uri)] = engine_session

    return engine_session


def init_hook(engine):
    function_map = {"sqlite": init_sqlite_hook}

    try:
        function_map[engine.dialect.name](engine)
    except KeyError:
        pass
    except Exception as e:
        logging.exception(e)
        logging.fatal(f"Can't initialize database.")


def init_sqlite_hook(engine):
    # enable the WAL
    engine.execute("PRAGMA journal_mode=WAL")


class DB:
    @classmethod
    def ensure_schema_exists(cls, engine, schema_name, grant_roles=()):
        """
        Make sure that the given schema_name exists in the database
        If not, create it

        :param db_conn: psycopg2 database connection
        :param schema_name: database schema
        """
        schema_identifier = schema_name
        group_identifiers = ",".join(grant_roles)

        create_schema = text(f"CREATE SCHEMA IF NOT EXISTS {schema_identifier}")
        grant_select_schema = text(
            f"ALTER DEFAULT PRIVILEGES IN SCHEMA {schema_identifier} GRANT SELECT ON TABLES TO {group_identifiers}"
        )
        grant_usage_schema = text(
            f"GRANT USAGE ON SCHEMA {schema_identifier} TO {group_identifiers}"
        )

        with engine.connect() as conn, conn.begin():
            conn.execute(create_schema)
            if grant_roles:
                conn.execute(grant_select_schema)
                conn.execute(grant_usage_schema)

        logging.info("Schema {} has been created successfully.".format(schema_name))
        for role in grant_roles:
            logging.info("Usage has been granted for role: {}.".format(role))



import json
from .utils import enforce_secure_filename

from meltano.core.project import Project


class ReposHelper:
    def get_topic(self, namespace, topic_name):
        project = Project.find()
        filename = enforce_secure_filename(f"{topic_name}.topic.m5oc")
        topic = project.run_dir("models", namespace, filename)

        with topic.open() as f:
            topic = json.load(f)

        return topic



import asyncio
import click
import logging
import os
import secrets
import signal
import subprocess
from click_default_group import DefaultGroup

from . import cli
from .params import project
from meltano.core.config_service import ConfigService
from meltano.core.plugin.error import PluginMissingError
from meltano.core.db import project_engine
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.utils import truthy
from meltano.core.migration_service import MigrationService
from meltano.api.workers import (
    MeltanoCompilerWorker,
    AirflowWorker,
    APIWorker,
    UIAvailableWorker,
)


logger = logging.getLogger(__name__)


def start_workers(workers):
    def stop_all():
        logger.info("Stopping all background workers...")
        for worker in workers:
            worker.stop()

    # start all workers
    for worker in workers:
        worker.start()

    return stop_all


@cli.group(cls=DefaultGroup, default="start", default_if_no_args=True)
@project(migrate=True)
@click.pass_context
def ui(ctx, project):
    ctx.obj["project"] = project


@ui.command()
@click.option("--reload", is_flag=True, default=False)
@click.option(
    "--bind-port",
    default=5000,
    help="Port to run webserver on",
    envvar="MELTANO_API_PORT",
    type=int,
)
@click.option(
    "--bind",
    default="0.0.0.0",
    help="The hostname (or IP address) to bind on",
    envvar="MELTANO_API_HOSTNAME",
)
@click.pass_context
def start(ctx, reload, bind_port, bind):
    project = ctx.obj["project"]
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_ui()

    workers = []

    if not truthy(os.getenv("MELTANO_DISABLE_AIRFLOW", False)):
        try:
            config_service = ConfigService(project)
            config_service.find_plugin("airflow")

            workers.append(AirflowWorker(project))
        except PluginMissingError:
            pass

    try:
        compiler_worker = MeltanoCompilerWorker(project)
        compiler_worker.compiler.compile()
        workers.append(compiler_worker)
    except Exception as e:
        logger.error(f"Initial compilation failed: {e}")

    workers.append(UIAvailableWorker("http://localhost:{bind_port}"))
    workers.append(
        APIWorker(
            project,
            f"{bind}:{bind_port}",
            reload=reload or os.getenv("FLASK_ENV") == "development",
        )
    )

    cleanup = start_workers(workers)

    def handle_terminate(signal, frame):
        cleanup()

    signal.signal(signal.SIGTERM, handle_terminate)
    logger.info("All workers started.")


@ui.command()
@click.argument("server_name")
@click.option("--bits", default=256)
@click.pass_context
def setup(ctx, server_name, **flags):
    """
    Generates the `ui.cfg` file to keep the server secrets keys.
    """
    project = ctx.obj["project"]
    ui_file_path = project.root_dir("ui.cfg")

    if ui_file_path.exists():
        logging.critical(
            f"Found secrets in file `{ui_file_path}`, please delete this file to regenerate the secrets."
        )
        raise click.Abort()

    generate_secret = lambda: secrets.token_hex(int(flags["bits"] / 8))  # in bytes

    config = {
        "SERVER_NAME": server_name,
        "SECRET_KEY": generate_secret(),
        "SECURITY_PASSWORD_SALT": generate_secret(),
    }

    # Flask doesn't support `configparser` or any other configuration format
    # than plain Python files.
    #
    # Luckily the format is trivial to generate
    with ui_file_path.open("w") as f:
        for k, v in config.items():
            f.write(f'{k} = "{v}"\n')



import sys
import click
import logging
import warnings

# disable the psycopg2 warning
# this needs to run before `psycopg2` is imported
warnings.filterwarnings("ignore", category=UserWarning, module="psycopg2")

import meltano
from meltano.core.project import Project, ProjectNotFound
from meltano.core.behavior.versioned import IncompatibleVersionError
from meltano.core.logging import setup_logging, LEVELS


logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()), default="info")
@click.option("-v", "--verbose", count=True)
@click.version_option(version=meltano.__version__, prog_name="meltano")
@click.pass_context
def cli(ctx, log_level, verbose):
    """
    Get help at https://www.meltano.com/docs/command-line-interface.html
    """
    setup_logging(log_level=log_level)

    ctx.ensure_object(dict)
    ctx.obj["verbosity"] = verbose

    try:
        ctx.obj["project"] = Project.find()
    except ProjectNotFound as err:
        ctx.obj["project"] = None
    except IncompatibleVersionError as err:
        click.secho(
            "This meltano project is incompatible with this version of `meltano`.",
            fg="yellow",
        )
        click.echo(
            "Visit http://meltano.com/docs/installation.html#upgrading-meltano-version for more details."
        )
        sys.exit(3)
